/*
 * taskview-monitor.c
 *
 * TaskviewMonitor
 *
 * Copyright © 2010 Salomon Sickert
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the licence or (at
 * your option) any later version.
 *
 * See the included COPYING file for more information.
 *
 * Authors:
 *    Salomon Sickert <sickert@in.tum.de>
 */

#include "taskview.h"
#include "taskview-monitor.h"

#define TASKVIEW_MONITOR_GET_PRIVATE(obj)	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), TASKVIEW_TYPE_MONITOR, TaskviewMonitorPrivate))

G_DEFINE_TYPE (TaskviewMonitor, taskview_monitor, G_TYPE_OBJECT);

struct _TaskviewMonitorPrivate
{
	gboolean claim_name;
  GDBusConnection *connection;
  GHashTable *app_proxies;
  guint subscription_id;
  guint owner_id;
};

enum /* signals */
{
  WORKER_APPEARED,
  WORKER_DISAPPEARED,
  LAST_SIGNAL
};

static guint _signals[LAST_SIGNAL] = { 0, 0, };

/* ----------------------------------------------------------------------------------------------------
 * GObject Properties
 * ---------------------------------------------------------------------------------------------------- */

enum
{
  PROP_0,

  PROP_CLAIM_NAME
};

static void
taskview_monitor_set_property (GObject      *object,
                          guint         property_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
  TaskviewMonitor *self = TASKVIEW_MONITOR (object);

  if (property_id == PROP_CLAIM_NAME)
    self->priv->claim_name = g_value_get_boolean (value);
  else 
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
}

/* ----------------------------------------------------------------------------------------------------
 * Helper
 * ---------------------------------------------------------------------------------------------------- */

static gboolean
is_tv_app_name (const gchar *name)
{
  return ((g_str_has_prefix (name, TASKVIEW_BUS_PREFIX) && g_strcmp0 (name, TASKVIEW_MONITOR_BUS_NAME) != 0));
}

static void
create_proxy (TaskviewMonitor *self, const gchar *name)
{
  TaskviewApplicationProxy *proxy;
  proxy = g_object_new (TASKVIEW_TYPE_APPLICATION_PROXY,
                        "dbus-name", name,
                        "dbus-connection", g_object_ref (self->priv->connection),
                        NULL);

  g_hash_table_insert (self->priv->app_proxies, g_strdup (name), proxy);
  g_signal_emit_by_name (self, "worker-appeared", name);
}

static void
remove_proxy (TaskviewMonitor *self, const gchar *name)
{
  g_hash_table_remove (self->priv->app_proxies, name);
  g_signal_emit_by_name (self, "worker-disappeared", name);
}


/* ----------------------------------------------------------------------------------------------------
 * Signal Handler
 * ---------------------------------------------------------------------------------------------------- */

static void
on_name_owner_changed (GDBusConnection *connection,
                       const gchar *sender_name,
                       const gchar *object_path,
                       const gchar *interface_name,
                       const gchar *signal_name,
                       GVariant *parameters,
                       gpointer user_data)
{
  g_assert (TASKVIEW_IS_MONITOR (user_data));

  TaskviewMonitor *self;
  gchar *name;
  gchar *old_owner;
  gchar *new_owner;

  self = TASKVIEW_MONITOR (user_data);
  g_variant_get (parameters, "(sss)", &name, &old_owner, &new_owner);

  if (is_tv_app_name (name))
    {
      if (g_strcmp0 (old_owner, "") == 0)
        create_proxy (self, name);
      else if (g_strcmp0 (new_owner, "") == 0)
        remove_proxy (self, name);
      else
        g_debug ("%s, %s, %s", name, old_owner, new_owner);
    }
}

static void
on_name_acquired (GDBusConnection *connection,
                  const gchar     *name,
                  gpointer         user_data)
{
  g_print ("Acquired the name %s on the session bus\n", name);
}

static void
on_name_lost (GDBusConnection *connection,
              const gchar     *name,
              gpointer         user_data)
{
  g_print ("Lost the name %s on the session bus\n", name);
}

static void
on_list_names (GObject *source_object,
               GAsyncResult *res,
               gpointer user_data)
{
	TaskviewMonitor *self = TASKVIEW_MONITOR (user_data);
	GVariantIter iter;
	GVariant *result, *sarray, *child;
	const gchar *bus_name;
	GError *error = NULL;

	result = g_dbus_connection_call_finish (self->priv->connection, res, &error);

	if (result == NULL)
	{
		g_printerr ("Error calling ListNames: %s\n", error->message);
		g_error_free (error);
		return;
	}

	sarray = g_variant_get_child_value (result, 0);
	g_variant_iter_init (&iter, sarray);

	while ((child = g_variant_iter_next_value (&iter)))
		{
			bus_name = g_variant_get_string (child, NULL);

			if (is_tv_app_name (bus_name))
				create_proxy (self, bus_name);
			
			g_variant_unref (child);
		}

	g_variant_unref (sarray);
	g_variant_unref (result);
}

static void
on_bus_acquired (GObject *source_object,
                 GAsyncResult *res,
                 gpointer user_data)
{
	TaskviewMonitor *self = TASKVIEW_MONITOR (user_data);
	GDBusConnection *connection;
	GError *error = NULL;

	connection = g_bus_get_finish (res, &error);

	if (connection == NULL)
	{
		g_printerr ("Error creating proxy: %s\n", error->message);
		g_error_free (error);
		return;
	}

	self->priv->connection = connection;
	self->priv->subscription_id = g_dbus_connection_signal_subscribe (connection,
	                                                                  "org.freedesktop.DBus",
	                                                                  "org.freedesktop.DBus",
	                                                                  "NameOwnerChanged",
	                                                                  "/org/freedesktop/DBus",
	                                                                  NULL,
	                                                                  G_DBUS_SIGNAL_FLAGS_NONE,
	                                                                  on_name_owner_changed,
	                                                                  self,
	                                                                  NULL);
	g_dbus_connection_call (connection,
	                        "org.freedesktop.DBus",
	                        "/org/freedesktop/DBus",
	                        "org.freedesktop.DBus",
	                        "ListNames",
	                        NULL,
	                        G_VARIANT_TYPE ("(as)"),
	                        G_DBUS_CALL_FLAGS_NONE,
	                        -1,
	                        NULL,
	                        on_list_names,
	                        self);


	if (self->priv->claim_name)
		self->priv->owner_id = g_bus_own_name_on_connection (connection,
		                                                     TASKVIEW_MONITOR_BUS_NAME,
		                                                     G_BUS_NAME_OWNER_FLAGS_NONE,
		                                                     on_name_acquired,
		                                                     on_name_lost,
		                                                     self,
		                                                     NULL);
}



/* ----------------------------------------------------------------------------------------------------
 * Initialization
 * ---------------------------------------------------------------------------------------------------- */

static GObject*
taskview_monitor_constructor (GType type,
                              guint n_construct_params,
                              GObjectConstructParam *construct_params)
{
  static GObject *self = NULL;

  if (self == NULL)
    {
      self = G_OBJECT_CLASS (taskview_monitor_parent_class)->constructor (type, n_construct_params, construct_params);
      g_object_add_weak_pointer (self, (gpointer) &self);
      return self;
    }

  return g_object_ref (self);
}

static void
taskview_monitor_constructed (GObject *gobject)
{
	g_bus_get (G_BUS_TYPE_SESSION,
	           NULL,
             on_bus_acquired,
             gobject);
	
	if (G_OBJECT_CLASS (taskview_monitor_parent_class)->constructed != NULL)
	  G_OBJECT_CLASS (taskview_monitor_parent_class)->constructed (gobject);
}

static void
taskview_monitor_dispose (GObject *gobject)
{
  TaskviewMonitor *self = TASKVIEW_MONITOR (gobject);

  if (self->priv->app_proxies)
    {
      g_hash_table_unref (self->priv->app_proxies);
      self->priv->app_proxies = NULL;
    }

  if (self->priv->connection)
    {
      g_dbus_connection_signal_unsubscribe (self->priv->connection, self->priv->subscription_id);
      g_bus_unown_name (self->priv->owner_id);
      g_object_unref (self->priv->connection);
      self->priv->connection = NULL;
    }

  G_OBJECT_CLASS (taskview_monitor_parent_class)->dispose (gobject);
}

static void
taskview_monitor_finalize (GObject *gobject)
{
  //TaskviewMonitor *self = TASKVIEW_MONITOR (gobject);

  //g_free (???);

  G_OBJECT_CLASS (taskview_monitor_parent_class)->finalize (gobject);
}

static void
taskview_monitor_class_init (TaskviewMonitorClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  g_type_class_add_private (gobject_class, sizeof (TaskviewMonitorPrivate));
	
  gobject_class->constructor = taskview_monitor_constructor;
  gobject_class->dispose = taskview_monitor_dispose;
  //gobject_class->finalize = taskview_monitor_finalize;
  gobject_class->constructed = taskview_monitor_constructed;
  gobject_class->set_property = taskview_monitor_set_property;
  
  _signals[WORKER_APPEARED] = g_signal_new ("worker-appeared",
                              TASKVIEW_TYPE_MONITOR,
                              G_SIGNAL_RUN_LAST,
                              0,
                              NULL,
                              NULL,
                              g_cclosure_marshal_VOID__STRING,
                              G_TYPE_NONE /* return_type */,
                              1    /* n_params */,
                              G_TYPE_STRING /* param_types */);


  _signals[WORKER_DISAPPEARED] = g_signal_new ("worker-disappeared",
                                 TASKVIEW_TYPE_MONITOR,
                                 G_SIGNAL_RUN_LAST,
                                 0,
                                 NULL,
                                 NULL,
                                 g_cclosure_marshal_VOID__STRING,
                                 G_TYPE_NONE /* return_type */,
                                 1     /* n_params */,
                                 G_TYPE_STRING  /* param_types */);

  GParamSpec *pspec;
	
  /* Claim Name */
  pspec = g_param_spec_boolean ("claim-name",
                               	"Claim Name",
                               	"Wheter the monitor object should claim "TASKVIEW_MONITOR_BUS_NAME,
                               	FALSE,
                               	G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE | G_PARAM_STATIC_STRINGS);	
  g_object_class_install_property (gobject_class, PROP_CLAIM_NAME, pspec);
}

static void
taskview_monitor_init (TaskviewMonitor *self)
{
  self->priv = TASKVIEW_MONITOR_GET_PRIVATE (self);

  self->priv->app_proxies = g_hash_table_new_full (g_str_hash,
                            g_str_equal,
                            g_free,
                            g_object_unref);
}

const GHashTable *
taskview_monitor_get_app_proxies (TaskviewMonitor *monitor)
{
  g_return_val_if_fail (TASKVIEW_IS_MONITOR (monitor), NULL);
  return monitor->priv->app_proxies;
}

TaskviewApplicationProxy *
taskview_monitor_get_app_proxy (TaskviewMonitor *monitor, const gchar *name)
{
  g_return_val_if_fail (TASKVIEW_IS_MONITOR (monitor), NULL);
  return TASKVIEW_APPLICATION_PROXY (g_hash_table_lookup (monitor->priv->app_proxies, name));
}
